/* TroiaPreview
 * version 0.4 date 2004-09-13
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>

#include "Globals.h"
#include "Pixel.h"
#include "Frame.h"
#include "Animations.h"

//private global variables
double AnimInterval = 0.0; //animations stopped
double AnimLastStepTime; //time of last step
unsigned int AnimStepNo; //step of animation
unsigned int AnimPixelMap, AnimPreMap, AnimNoMap, AnimPostMap; //mappings methods

//map pixel to some value in range -1.0 ... 1.0
static double AnimStepPixelMap( tPixel3D Pixel, unsigned int Method )
{
  //which method to use
  switch( Method )
  {
    //some corners -1.0, some corners 0.0, some corners 1.0
    case 0:
      switch( Pixel.Wall )
      {
        case 1:
        case 3:
          return (double)Pixel.X / (double)(PixelX-1)
              + (double)Pixel.Y / (double)(PixelZ-1)
              - 1.0;
        case 2:
        case 4:
          return -(double)Pixel.X / (double)(PixelY-1)
              + (double)Pixel.Y / (double)(PixelZ-1);
        case 5:
          return -Abs( (double)Pixel.X / (double)(PixelX-1)
              + (double)Pixel.Y / (double)(PixelY-1)
              - 1.0 );
          break;
        case 6:
          return Abs( (double)Pixel.X / (double)(PixelX-1)
               - (double)Pixel.Y / (double)(PixelY-1) );
      }
      break;
    //corners 1.0, middles -1.0
    case 1:
      switch( Pixel.Wall )
      {
        case 1:
        case 3:
          return Abs( (double)Pixel.X / ((double)(PixelX-1) * 0.5) - 1.0 )
            + Abs( (double)Pixel.Y / ((double)(PixelZ-1) * 0.5) - 1.0 )
            - 1.0;
        case 2:
        case 4:
          return Abs( (double)Pixel.X / ((double)(PixelY-1) * 0.5) - 1.0 )
            + Abs( (double)Pixel.Y / ((double)(PixelZ-1) * 0.5) - 1.0 )
            - 1.0;
        case 5:
        case 6:
          return Abs( (double)Pixel.X / ((double)(PixelX-1) * 0.5) - 1.0 )
            + Abs( (double)Pixel.Y / ((double)(PixelY-1) * 0.5) - 1.0 )
            - 1.0;
      }
      break;
    //vertical edges -1.0 or 1.0
    case 2:
      switch( Pixel.Wall )
      {
        case 1:
        case 3:
        case 5:
        case 6:
          return 2.0 * (double)Pixel.X / (double)(PixelX-1) - 1.0;
        case 2:
        case 4:
          return -2.0 * (double)Pixel.X / (double)(PixelY-1) + 1.0;
      }
      break;
    //horizontal edges -1.0 or 1.0
    case 3:
      switch( Pixel.Wall )
      {
        case 1:
        case 2:
        case 3:
        case 4:
          return 2.0 * (double)Pixel.Y / (double)(PixelZ-1) - 1.0;
        case 5:
        case 6:
          return 2.0 * (double)Pixel.Y / (double)(PixelY-1) - 1.0;
      }
      break;
  }

  //default: always 0.0
  return 0.0;
}

//map value to another value
static double AnimStepMap( double Value, unsigned int Method )
{
  //which method to use
  switch( Method )
  {
    //plain
    case 0:
      return Value;
    //smooth modulo multiplicated with 5
    case 1:
      Value = DoubleModulo( Value * 5.0 + 1.0, 4.0 ) - 1.0;
      if( Value > 1.0 )
        Value = 2.0 - Value;
      return Value;
    //smooth modulo multiplicated with 9
    case 2:
      Value = DoubleModulo( Value * 9.0 + 1.0, 4.0 ) - 1.0;
      if( Value > 1.0 )
        Value = 2.0 - Value;
      return Value;
  }

  //default: plain
  return Value;
}

//map animation step number into value
static double AnimStepNoMap( double Value, unsigned int AnimStep, unsigned int Method )
{
  //which method to use
  switch( Method )
  {
    //do not use animation step
    case 0:
      return Value;
    //modulo mapping
    case 1:
      return DoubleModulo( Value + (double)(AnimStep & 31) * 0.125 + 1.0, 2.0 ) - 1.0;
    //smooth modulo mapping
    case 2:
      Value = DoubleModulo( Value + (double)(AnimStep & 31) * 0.125 + 1.0, 4.0 ) - 1.0;
      if( Value > 1.0 )
        Value = 2.0 - Value;
      return Value;
  }

  //default: do not use animation step
  return Value;
}

//callback for animation step
unsigned char AnimStepCallback( tPixel3D Pixel, void * pUnused )
{
  double V;

  //keep compiler happy
  pUnused = NULL;

  //map pixel to some value in range -1.0 ... 1.0
  V = AnimStepPixelMap( Pixel, AnimPixelMap );

  //map value to another value
  V = AnimStepMap( V, AnimPreMap );

  //map animation step number into value
  V = AnimStepNoMap( V, AnimStepNo, AnimNoMap );

  //map value to another value
  V = AnimStepMap( V, AnimPostMap );

  //return color according to value (-1.0=>0, 1.0=>255)
  return (unsigned char)(V * 127.5 + 127.5);
}

//do an animation step
static void AnimStep( )
{
  FrameColorByCallback( AnimStepCallback, NULL );
  AnimStepNo++;
}

//start simple animation
void AnimStart( unsigned int PixelMap, unsigned int PreMap,
          unsigned int NoMap, unsigned int PostMap, double Interval )
{
  AnimInterval = DoubleMinMax( Interval, AnimMinInterval, AnimMaxInterval );
  AnimLastStepTime = TimeGet( ) - AnimInterval;
  AnimStepNo = 0;
  AnimPixelMap = PixelMap;
  AnimPreMap = PreMap;
  AnimNoMap = NoMap;
  AnimPostMap = PostMap;
}

//stop simple animation
void AnimStop( )
{
  AnimInterval = 0.0;
}

//called periodically to do some work
void AnimTask( )
{
  double CurrentTime;

  //exit if animation not running
  if( AnimInterval < AnimMinInterval * 0.5 ) //use * 0.5 to get a safe decision threshold
    return;

  //get time
  CurrentTime = TimeGet( );

  //exit if interval not elapsed
  if( CurrentTime < AnimInterval + AnimLastStepTime )
    return;

  //do a step
  AnimStep( );

  //restart interval
  AnimLastStepTime = CurrentTime;
}
