/* TroiaPreview
 * version 0.4 date 2004-09-13
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Tools.h"
#include "Globals.h"
#include "Pixel.h"

//convert 3D pixel structure to text
//pTxt must point to at least 16 bytes
void Pixel3D2Txt( tPixel3D Pixel, char * pTxt )
{
  switch( Pixel.Wall )
  {
    case 1:
      sprintf( pTxt, "W1/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelX - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelZ - 1 ) );
      break;
    case 2:
      sprintf( pTxt, "W2/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelY - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelZ - 1 ) );
      break;
    case 3:
      sprintf( pTxt, "W3/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelX - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelZ - 1 ) );
      break;
    case 4:
      sprintf( pTxt, "W4/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelY - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelZ - 1 ) );
      break;
    case 5:
      sprintf( pTxt, "D/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelX - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelY - 1 ) );
      break;
    case 6:
      sprintf( pTxt, "B/%d/%d",
        ShortMinMax( Pixel.X, 0, PixelX - 1 ),
        ShortMinMax( Pixel.Y, 0, PixelY - 1 ) );
      break;
    default:
      strcpy( pTxt, "?/?/?" );
  }
}

//convert text to 3D pixel structure
tPixel3D Txt2Pixel3D( char * pTxt )
{
  char * pChr1, * pChr2;
  short MaxX, MaxY;
  tPixel3D Pixel;

  //invalid pixel by default
  Pixel.Wall = 0;
  Pixel.X = 0;
  Pixel.Y = 0;

  //search slashes
  pChr1 = strchr( pTxt, '/' );
  if( pChr1 == NULL )
    return Pixel;
  pChr2 = strchr( pChr1 + 1, '/' );
  if( pChr2 == NULL )
    return Pixel;

  //get pixel type
  MaxX = 0;
  MaxY = 0;
  for( ; Pixel.Wall == 0 && *pTxt != '/'; pTxt++ )
  {
    switch( *pTxt )
    {
      //long walls
      case '1':
      case '3':
        Pixel.Wall = *pTxt - '0';
        MaxX = PixelX - 1;
        MaxY = PixelZ - 1;
        break;
      //short walls
      case '2':
      case '4':
        Pixel.Wall = *pTxt - '0';
        MaxX = PixelX - 1;
        MaxY = PixelZ - 1;
        break;
      //ceiling
      case 'C':
      case 'c':
      case 'D':
      case 'd':
        Pixel.Wall = 5;
        MaxX = PixelX - 1;
        MaxY = PixelY - 1;
        break;
      //floor
      case 'B':
      case 'b':
      case 'F':
      case 'f':
        Pixel.Wall = 6;
        MaxX = PixelX - 1;
        MaxY = PixelY - 1;
        break;
    }
  }
  if( Pixel.Wall == 0 )
    return Pixel;

  //get numbers
  Pixel.X = ShortMinMax( (short)atoi( pChr1 + 1 ), 0, MaxX );
  Pixel.Y = ShortMinMax( (short)atoi( pChr2 + 1 ), 0, MaxY );

  return Pixel;
}

//get a random pixel
tPixel3D PixelRandom( )
{
  tPixel3D Pixel;
  Pixel.Wall = (unsigned char)(rand( ) % 6 + 1);
  switch( Pixel.Wall )
  {
    case 1:
    case 3:
      Pixel.X = (unsigned short)(rand( ) % PixelX);
      Pixel.Y = (unsigned short)(rand( ) % PixelZ);
      break;
    case 2:
    case 4:
      Pixel.X = (unsigned short)(rand( ) % PixelY);
      Pixel.Y = (unsigned short)(rand( ) % PixelZ);
      break;
    case 5:
    case 6:
      Pixel.X = (unsigned short)(rand( ) % PixelX);
      Pixel.Y = (unsigned short)(rand( ) % PixelY);
      break;
  }
  return Pixel;
}

//convert direction to text
//direction is encoded as the number on the numpad (e.g. 8 for up)
//pTxt must point to at least 16 bytes
void Dir2Txt( unsigned char Dir, char * pTxt )
{
  switch( Dir )
  {
    case 1: strcpy( pTxt, "downleft" ); break;
    case 2: strcpy( pTxt, "down" ); break;
    case 3: strcpy( pTxt, "downright" ); break;
    case 4: strcpy( pTxt, "left" ); break;
    case 6: strcpy( pTxt, "right" ); break;
    case 7: strcpy( pTxt, "upleft" ); break;
    case 8: strcpy( pTxt, "up" ); break;
    case 9: strcpy( pTxt, "upright" ); break;
    default: strcpy( pTxt, "???" );
  }
}

//convert text to direction
//direction is encoded as the number on the numpad (e.g. 8 for up)
unsigned char Txt2Dir( char * pTxt )
{
  if( strncasecmp( pTxt, "downleft", 8 ) == 0 )
    return 1;
  if( strncasecmp( pTxt, "leftdown", 8 ) == 0 )
    return 1;
  if( strncasecmp( pTxt, "downright", 9 ) == 0 )
    return 3;
  if( strncasecmp( pTxt, "rightdown", 9 ) == 0 )
    return 3;
  if( strncasecmp( pTxt, "upleft", 6 ) == 0 )
    return 7;
  if( strncasecmp( pTxt, "leftup", 6 ) == 0 )
    return 7;
  if( strncasecmp( pTxt, "upright", 7 ) == 0 )
    return 9;
  if( strncasecmp( pTxt, "rightup", 7 ) == 0 )
    return 9;
  if( strncasecmp( pTxt, "down", 4 ) == 0 )
    return 2;
  if( strncasecmp( pTxt, "left", 4 ) == 0 )
    return 4;
  if( strncasecmp( pTxt, "right", 5 ) == 0 )
    return 6;
  if( strncasecmp( pTxt, "up", 2 ) == 0 )
    return 8;
  return 0;
}

//get a random direction
unsigned char DirRandom( )
{
  unsigned char R;
  R = (unsigned char)(rand( ) & 7);
  R++;
  if( R >= 5 )
    R++;
  return R;
}

//rotate a direction
//direction is encoded as the number on the numpad (e.g. 8 for up)
//angle +1 rotates 45 degrees clockwise
unsigned char DirRotate( unsigned char Dir, int Angle )
{
  //simplify angle
  if( Angle < 0 )
    Angle = 8 - ((-Angle) % 8);
  Angle %= 8;

  //rotate clockwise
  for( ; Angle >0; Angle-- )
  {
    switch( Dir )
    {
      case 1: Dir = 4; break;
      case 2: Dir = 1; break;
      case 3: Dir = 2; break;
      case 4: Dir = 7; break;
      case 6: Dir = 3; break;
      case 7: Dir = 8; break;
      case 8: Dir = 9; break;
      case 9: Dir = 6; break;
      default: return 0; //invalid direction
    }
  }

  return Dir;
}

//move from one pixel to a neighbour pixel
//direction is encoded as the number on the numpad (e.g. 8 for up)
//the new direction (to follow a straight line) is also returned
void Pixel3DMove( tPixel3D Src, unsigned char Dir, tPixel3D * pDest, unsigned char * pNewDir )
{
  //default: invaild destination pixel, invalid new direction
  pDest->Wall = 0;
  pDest->X = 0;
  pDest->Y = 0;
  *pNewDir = 0;

  switch( Src.Wall )
  {
    //wall 1
    case 1:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 1;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = Src.X - 1;
              pDest->Y = 0;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y < PixelZ - 1 )
          {
            pDest->Wall = 1;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 6;
            pDest->X = Src.X;
            pDest->Y = 0;
            *pNewDir = 2;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 1;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = Src.X + 1;
              pDest->Y = 0;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 2;
              pDest->X = 0;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 1;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 4;
            pDest->X = PixelY - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelX - 1 )
          {
            pDest->Wall = 1;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 2;
            pDest->X = 0;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 1;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = Src.X - 1;
              pDest->Y = 0;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y > 0 )
          {
            pDest->Wall = 1;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 5;
            pDest->X = Src.X;
            pDest->Y = 0;
            *pNewDir = 8;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 1;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = Src.X + 1;
              pDest->Y = 0;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = 0;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
          }
          break;
      } //switch( Dir )
      break;
    //wall 2
    case 2:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 2;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = PixelX - 1;
              pDest->Y = Src.X - 1;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 1;
              pDest->X = PixelX - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y < PixelZ - 1 )
          {
            pDest->Wall = 2;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 6;
            pDest->X = PixelX - 1;
            pDest->Y = Src.X;
            *pNewDir = 4;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelY - 1 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 2;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = PixelX - 1;
              pDest->Y = Src.X + 1;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 3;
              pDest->X = 0;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 2;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 1;
            pDest->X = PixelX - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelY - 1 )
          {
            pDest->Wall = 2;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 3;
            pDest->X = 0;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = PixelX - 1;
              pDest->Y = Src.X - 1;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 1;
              pDest->X = PixelX - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y > 0 )
          {
            pDest->Wall = 2;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 5;
            pDest->X = PixelX - 1;
            pDest->Y = Src.X;
            *pNewDir = 4;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelY - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = PixelX - 1;
              pDest->Y = Src.X + 1;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 3;
              pDest->X = 0;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
          }
          break;
      } //switch( Dir )
      break;
    //wall 3
    case 3:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 3;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = PixelX - Src.X;
              pDest->Y = PixelY - 1;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 2;
              pDest->X = PixelY - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y < PixelZ - 1 )
          {
            pDest->Wall = 3;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 6;
            pDest->X = PixelX - 1 - Src.X;
            pDest->Y = PixelY - 1;
            *pNewDir = 8;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 3;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = PixelX - 2 - Src.X;
              pDest->Y = PixelY - 1;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 4;
              pDest->X = 0;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 3;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 2;
            pDest->X = PixelY - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelX - 1 )
          {
            pDest->Wall = 3;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 4;
            pDest->X = 0;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 3;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = PixelX - Src.X;
              pDest->Y = PixelY - 1;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = PixelY - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y > 0 )
          {
            pDest->Wall = 3;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 5;
            pDest->X = PixelX - 1 - Src.X;
            pDest->Y = PixelY - 1;
            *pNewDir = 2;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 3;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = PixelX - 2 - Src.X;
              pDest->Y = PixelY - 1;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = 0;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
          }
          break;
      } //switch( Dir )
      break;
    //wall 4
    case 4:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 4;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = 0;
              pDest->Y = PixelY - Src.X;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 3;
              pDest->X = PixelX - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y < PixelZ - 1 )
          {
            pDest->Wall = 4;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 6;
            pDest->X = 0;
            pDest->Y = PixelY - 1 - Src.X;
            *pNewDir = 6;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelY - 1 )
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 4;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 6;
              pDest->X = 0;
              pDest->Y = PixelY - 2 - Src.X;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y < PixelZ - 1 )
            {
              pDest->Wall = 1;
              pDest->X = 0;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 4;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 3;
            pDest->X = PixelX - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelY - 1 )
          {
            pDest->Wall = 4;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 1;
            pDest->X = 0;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = 0;
              pDest->Y = PixelY - Src.X;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 3;
              pDest->X = PixelX - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y > 0 )
          {
            pDest->Wall = 4;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 5;
            pDest->X = 0;
            pDest->Y = PixelY - 1 - Src.X;
            *pNewDir = 6;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelY - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 5;
              pDest->X = 0;
              pDest->Y = PixelY - 2 - Src.X;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 1;
              pDest->X = 0;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
          }
          break;
      } //switch( Dir )
      break;
    //ceiling
    case 5:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 5;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 1;
              pDest->X = Src.X - 1;
              pDest->Y = 0;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - Src.Y;
              pDest->Y = 0;
              *pNewDir = 3;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y > 0 )
          {
            pDest->Wall = 5;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 1;
            pDest->X = Src.X;
            pDest->Y = 0;
            *pNewDir = 2;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 5;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 1;
              pDest->X = Src.X + 1;
              pDest->Y = 0;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = Src.Y - 1;
              pDest->Y = 0;
              *pNewDir = 1;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 5;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 4;
            pDest->X = PixelY - 1 - Src.Y;
            pDest->Y = 0;
            *pNewDir = 2;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelX - 1 )
          {
            pDest->Wall = 5;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 2;
            pDest->X = Src.Y;
            pDest->Y = 0;
            *pNewDir = 2;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 5;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 3;
              pDest->X = PixelX - Src.X;
              pDest->Y = 0;
              *pNewDir = 3;
            }
          }
          else
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - 2 - Src.Y;
              pDest->Y = 0;
              *pNewDir = 1;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y < PixelY - 1 )
          {
            pDest->Wall = 5;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 3;
            pDest->X = PixelX - 1 - Src.X;
            pDest->Y = 0;
            *pNewDir = 2;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 5;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 3;
              pDest->X = PixelX - 2 - Src.X;
              pDest->Y = 0;
              *pNewDir = 1;
            }
          }
          else
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 2;
              pDest->X = Src.Y + 1;
              pDest->Y = 0;
              *pNewDir = 3;
            }
          }
          break;
      }
      break;
    //floor
    case 6:
      switch( Dir )
      {
        //down left
        case 1:
          if( Src.X > 0 )
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 6;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 1;
            }
            else
            {
              pDest->Wall = 3;
              pDest->X = PixelX - Src.X;
              pDest->Y = PixelZ - 1;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - 2 - Src.Y;
              pDest->Y = PixelZ - 1;
              *pNewDir = 7;
            }
          }
          break;
        //down
        case 2:
          if( Src.Y < PixelY - 1 )
          {
            pDest->Wall = 6;
            pDest->X = Src.X;
            pDest->Y = Src.Y + 1;
            *pNewDir = 2;
          }
          else
          {
            pDest->Wall = 3;
            pDest->X = PixelX - 1 - Src.X;
            pDest->Y = PixelZ - 1;
            *pNewDir = 8;
          }
          break;
        //down right
        case 3:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 6;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y + 1;
              *pNewDir = 3;
            }
            else
            {
              pDest->Wall = 3;
              pDest->X = PixelX - 2 - Src.X;
              pDest->Y = PixelZ - 1;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y < PixelY - 1 )
            {
              pDest->Wall = 2;
              pDest->X = Src.Y + 1;
              pDest->Y = PixelZ - 1;
              *pNewDir = 9;
            }
          }
          break;
        //left
        case 4:
          if( Src.X > 0 )
          {
            pDest->Wall = 6;
            pDest->X = Src.X - 1;
            pDest->Y = Src.Y;
            *pNewDir = 4;
          }
          else
          {
            pDest->Wall = 4;
            pDest->X = PixelY - 1 - Src.Y;
            pDest->Y = PixelZ - 1;
            *pNewDir = 8;
          }
          break;
        //right
        case 6:
          if( Src.X < PixelX - 1 )
          {
            pDest->Wall = 6;
            pDest->X = Src.X + 1;
            pDest->Y = Src.Y;
            *pNewDir = 6;
          }
          else
          {
            pDest->Wall = 2;
            pDest->X = Src.Y;
            pDest->Y = PixelZ - 1;
            *pNewDir = 8;
          }
          break;
        //up left
        case 7:
          if( Src.X > 0 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 6;
              pDest->X = Src.X - 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 7;
            }
            else
            {
              pDest->Wall = 1;
              pDest->X = Src.X - 1;
              pDest->Y = PixelZ - 1;
              *pNewDir = 7;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 4;
              pDest->X = PixelY - Src.Y;
              pDest->Y = PixelZ - 1;
              *pNewDir = 9;
            }
          }
          break;
        //up
        case 8:
          if( Src.Y > 0 )
          {
            pDest->Wall = 6;
            pDest->X = Src.X;
            pDest->Y = Src.Y - 1;
            *pNewDir = 8;
          }
          else
          {
            pDest->Wall = 1;
            pDest->X = Src.X;
            pDest->Y = PixelZ - 1;
            *pNewDir = 8;
          }
          break;
        //up right
        case 9:
          if( Src.X < PixelX - 1 )
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 6;
              pDest->X = Src.X + 1;
              pDest->Y = Src.Y - 1;
              *pNewDir = 9;
            }
            else
            {
              pDest->Wall = 1;
              pDest->X = Src.X + 1;
              pDest->Y = PixelZ - 1;
              *pNewDir = 9;
            }
          }
          else
          {
            if( Src.Y > 0 )
            {
              pDest->Wall = 2;
              pDest->X = Src.Y - 1;
              pDest->Y = PixelZ - 1;
              *pNewDir = 7;
            }
          }
          break;
      } //switch( Dir )
      break;
  } //switch( Src.Wall )
}

